###############################################################
#                                                             #
# Template code to compute stresses in a half plane subject   #
# to a distributed normal and shear loads on its surface.     #
# The latter are given by Cattaneo-Mindlin solution for       #
# partial slip between two cylinders.                         #
#                                                             #
# After computing stress state, the code finds principal      #
# stresses and plots them.                                    #
#                                                             #
# "Contact Mechanics and Elements of Tribology" (CMET) course #
#                                                             #
# V.A. Yastrebov, CNRS                                        #
# MINES Paris - PSL, Centre des materiaux                     #
# Feb 2021 - Jan 2024                                         #
#                                                             #
# Licence: BSD 3-clause                                       #
#                                                             #
###############################################################

import numpy as np
import scipy.integrate as integrate
import matplotlib.pyplot as plt
import matplotlib 
from matplotlib import cm
import numba

matplotlib.rcParams.update(matplotlib.rcParamsDefault)
matplotlib.rcParams['figure.figsize'] = 10,8
matplotlib.rcParams.update({'font.size': 14})

# Pressure distribution
def pres(x):
  if abs(x) <= a:
    return p0 * np.sqrt(1-(x/a)**2) 
  else:
    return 0.
  
# Shear distribution
def shear(x):
  if abs(x) <= c:
    return abs(p0)  * friction * ( np.sqrt(1-(x/a)**2) - np.sqrt((c/a)**2 - (x/a)**2))
  elif abs(x) <= a:
    return abs(p0) * friction * np.sqrt(1-(x/a)**2)
  else:
    return 0.
  
# Integrands
def fxx(x,y,s):
  return - 2 * pres(s)/np.pi * y *(x-s)**2 / ((x-s)**2 + y**2)**2 - 2* shear(s) * (x - s)**3 / (np.pi * ((x-s)**2 + y**2)**2)
def fyy(x,y,s):
  return - 2 * pres(s)/np.pi * y**3 / ((x-s)**2 + y**2)**2 - 2* shear(s) * y**2 * (x - s) / (np.pi * ((x-s)**2 + y**2)**2)
def fxy(x,y,s):
  return - 2 * pres(s)/np.pi * y**2 *(x-s) / ((x-s)**2 + y**2)**2 - 2* shear(s) * y * (x - s)**2 / (np.pi * ((x-s)**2 + y**2)**2)
# Integrate stress components
def get_stress(L,W,Nx,Ny):
  dx = L/Nx
  dy = W/Ny
  for i in range(Nx):
    x=-L/2.+ dx*(i+1/2)
    for j in range(Ny): 
      y = -W + dy*(j+1/2.)
      sig_xx[i][j] = integrate.quad(lambda s: fxx(x,y,s), -a, a)[0]
      sig_yy[i][j] = integrate.quad(lambda s: fyy(x,y,s), -a, a)[0]
      sig_xy[i][j] = integrate.quad(lambda s: fxy(x,y,s), -a, a)[0]
      sig_zz[i][j] = v * (sig_xx[i][j] + sig_yy[i][j])
  return sig_xx,sig_yy,sig_xy,sig_zz

# Function to plot stress fields
def PlotField(X: np.dtype(float),Y: np.dtype(float),Z: np.dtype(float),label: str,filename: str) -> None:
  fig,ax = plt.subplots()
  # ax.set_aspect(1)
  plt.xlabel("x/a")
  plt.ylabel("y/a")

  lev = np.linspace(np.min(Z/p0),np.max(Z/p0),NumColors)

  cs = ax.contourf(X/a, Y/a, Z.T/p0, levels=lev, cmap=cm.rainbow,origin="lower")
  ax.contour(X/a,Y/a,Z.T/p0,levels=lev,colors="k",origin="lower")
  cbar = fig.colorbar(cs)
  cbar.ax.set_ylabel(label)
  fig.savefig(filename)
  plt.show()

# Contact pressure is distributed from -a to a
friction = 0.5
# Poisson's ratio
E = 115e9 # GPa
v = 0.33
# Pressure constant
P = 1e6 # N
R = 1e-2 # m
Estar = E/(1-v**2)/2
a = ((4*P*R)/(np.pi*Estar))**0.5
c = 0.5*a
p0 = - 2*P/(np.pi*a)

Shear = np.vectorize(shear)
Pres = np.vectorize(pres)

x = np.linspace(-a,a,500)
fig,ax = plt.subplots()
plt.xlabel("x/a")
plt.ylabel("Surface tractions, $p/|p_0|$, $q/|p_0|$")
plt.ylim(0,1.1)
plt.grid()
plt.plot(x, Shear(x)/abs(p0),"-",color="r",lw=3,label="Shear")
plt.plot(x,-Pres(x)/abs(p0),"-",color="b",lw=3,label="Pressure")
plt.plot(x,-friction * Pres(x)/abs(p0),"--",color="k",label="Pressure * $\mu$")
plt.legend(loc="best")
fig.savefig("Tractions.pdf")
plt.show()


L = 4*a
Nx = 40
W = 1*a
Ny = 20
dx = L/Nx
dy = W/Ny
# Stress arrays
sig_xx = np.zeros((Nx,Ny))
sig_yy = np.zeros((Nx,Ny))
sig_xy = np.zeros((Nx,Ny))
sig_zz = np.zeros((Nx,Ny))
sig_vm = np.zeros((Nx,Ny)) #von Mises stress

sig_xx,sig_yy,sig_xy,sig_zz = get_stress(L,W,Nx,Ny)
sig_vm = np.sqrt(0.5*((sig_xx-sig_yy)**2 + (sig_xx - sig_zz)**2 + (sig_yy - sig_zz)**2 + 6*sig_xy**2))


# Create mesh grid (X,Y)
x = np.linspace(-L/2. + dx/2., L/2. - dx/2., Nx)
y = np.linspace(-W + dy/2., -dy/2., Ny)
X, Y = np.meshgrid(x, y)
# Number of isolines
NumColors = 11

# Plot results
# PlotField(X,Y,sig_xx,"sigma_xx/p0","sigma_xx.pdf")
# PlotField(X,Y,sig_yy,"sigma_yy/p0","sigma_yy.pdf")
# PlotField(X,Y,sig_xy,"sigma_xy/p0","sigma_xy.pdf")
# PlotField(X,Y,sig_vm,"sigma_vm/p0","sigma_vm.pdf")

# Compute principal stresses
# We are interested in the inplane eigenvalues and eigenvectors of the stress tensor
# sigma_zz always represent an eigenvalue with its eigenvector (0,0,1) because
#         (sxx sxy  0 )
# sigma = (sxy syy  0 )
#         ( 0   0  szz)
stress_tensor = np.zeros((Nx,Ny,2,2))
stress_tensor[:,:,0,0] = sig_xx
stress_tensor[:,:,1,1] = sig_yy
stress_tensor[:,:,0,1] = sig_xy
stress_tensor[:,:,1,0] = sig_xy

eigenvalues, eigen_directions = np.linalg.eig(stress_tensor)

# Find indices of the max and min eigenvalues
max_indices = np.argmax(eigenvalues, axis=-1)
min_indices = np.argmin(eigenvalues, axis=-1)

# Allocate arrays for principal stresses and directions
sig_1 = np.zeros((Nx, Ny))
sig_2 = np.zeros((Nx, Ny))
dir_1 = np.zeros((Nx, Ny, 2))
dir_2 = np.zeros((Nx, Ny, 2))

# Extract principal stresses and directions
for i in range(2):
    sig_1[max_indices == i] = eigenvalues[max_indices == i, i]
    sig_2[min_indices == i] = eigenvalues[min_indices == i, i]
    dir_1[max_indices == i] = eigen_directions[max_indices == i, i, :]
    dir_2[min_indices == i] = eigen_directions[min_indices == i, i, :]

# Calculate max shear stress
max_shear = (sig_1 - sig_2) / 2

sig_1 = sig_1.T
sig_2 = sig_2.T
max_shear = max_shear.T
dir_1 = dir_1.T
dir_2 = dir_2.T

U1 = dir_1[0,:,:]
V1 = dir_1[1,:,:]

U2 = dir_2[0,:,:]
V2 = dir_2[1,:,:]

# Plot principal stresses
plt.rcParams['figure.figsize'] = 10,5
fig,ax = plt.subplots()
# ax.set_aspect(1)
plt.xlabel("x/a")
plt.ylabel("y/a")
cs = ax.contourf(X/a, Y/a, sig_1/abs(p0), levels=NumColors, cmap=cm.rainbow)
ax.contour(X/a,Y/a,sig_1/abs(p0),levels=NumColors,colors="k")
plt.quiver(X/a,Y/a,U1,V1,headlength=0, headwidth=0, headaxislength=0)
cbar = fig.colorbar(cs)
cbar.ax.set_ylabel("sigma_1/p0")
fig.savefig("sigma_1.pdf")
plt.show()

# Plot maximum shear stress 
fig,ax = plt.subplots()
# ax.set_aspect(1)
plt.xlabel("x/a")
plt.ylabel("y/a")
cs = ax.contourf(X/a, Y/a, max_shear/abs(p0), levels=NumColors, cmap=cm.rainbow)
ax.contour(X/a,Y/a,max_shear/abs(p0),levels=NumColors,colors="k")
plt.quiver(X/a,Y/a,(U1-U2)/2.,(V1-V2)/2.,headlength=0, headwidth=0, headaxislength=0)
cbar = fig.colorbar(cs)
cbar.ax.set_ylabel("max shear stress/p0")
fig.savefig("max_shear.pdf")
plt.show()

# Plot angle for the maximum shear stress
fig,ax = plt.subplots()
# ax.set_aspect(1)
plt.xlabel("x/a")
plt.ylabel("y/a")
cs = ax.contourf(X/a, Y/a, -180/np.pi*np.arctan((V1-V2)/(U1-U2)), levels=NumColors, cmap=cm.rainbow)
# ax.contour(X/a,Y/a,-180/np.pi*np.arctan((V1-V2)/(U1-U2)),levels=NumColors,colors="k")
cbar = fig.colorbar(cs)
cbar.ax.set_ylabel("Maximum shear stress angle, deg")
fig.savefig("MaxShearAngle.pdf")
plt.show()